// Copyright (C) 2017 by Pedro Mendes, Virginia Tech Intellectual
// Properties, Inc., University of Heidelberg, and University of
// of Connecticut School of Medicine.
// All rights reserved.

// Copyright (C) 2010 - 2016 by Pedro Mendes, Virginia Tech Intellectual
// Properties, Inc., University of Heidelberg, and The University
// of Manchester.
// All rights reserved.

// Copyright (C) 2009 by Pedro Mendes, Virginia Tech Intellectual
// Properties, Inc., EML Research, gGmbH, University of Heidelberg,
// and The University of Manchester.
// All rights reserved.

#include "CQReportsWidget.h"

#include <QHeaderView>
#include <QClipboard>
#include <QKeyEvent>

#include "copasi.h"

#include "qtUtilities.h"
#include "CQMessageBox.h"

#include "model/CModel.h"
#include "CopasiDataModel/CDataModel.h"
#include "copasi/core/CRootContainer.h"
#include "report/CReportDefinitionVector.h"

/*
 *  Constructs a CQReportsWidget which is a child of 'parent', with the
 *  name 'name'.'
 */
CQReportsWidget::CQReportsWidget(QWidget *parent, const char *name)
  : CopasiWidget(parent, name)
{
  setupUi(this);
  //Create Source Data Model.
  mpReportDM = new CQReportDM(this, mpDataModel);
  //Create the Proxy Model for sorting/filtering and set its properties.
  mpProxyModel = new CQSortFilterProxyModel();
  mpProxyModel->setSortCaseSensitivity(Qt::CaseInsensitive);
  mpProxyModel->setFilterKeyColumn(-1);
#if QT_VERSION >= 0x050000
  mpTblReports->verticalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
  mpTblReports->verticalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif
  mpTblReports->verticalHeader()->hide();
  mpTblReports->sortByColumn(COL_ROW_NUMBER, Qt::AscendingOrder);
  setFramework(mFramework);
  // Connect the table widget
  connect(mpReportDM, SIGNAL(notifyGUI(ListViews::ObjectType, ListViews::Action, const std::string)),
          this, SLOT(protectedNotify(ListViews::ObjectType, ListViews::Action, const std::string)));
  connect(mpReportDM, SIGNAL(dataChanged(const QModelIndex &, const QModelIndex &)),
          this, SLOT(dataChanged(const QModelIndex &, const QModelIndex &)));
  connect(mpLEFilter, SIGNAL(textChanged(const QString &)),
          this, SLOT(slotFilterChanged()));
}

/*
 *  Destroys the object and frees any allocated resources
 */
CQReportsWidget::~CQReportsWidget()
{
  pdelete(mpProxyModel);
  pdelete(mpReportDM);
  // no need to delete child widgets, Qt does it all for us
}

void CQReportsWidget::slotBtnNewClicked()
{
  mpReportDM->insertRow(mpReportDM->rowCount() - 1, QModelIndex());
  updateDeleteBtns();
}

void CQReportsWidget::slotBtnDeleteClicked()
{
  if (mpTblReports->hasFocus())
    {deleteSelectedReports();}

  updateDeleteBtns();
}

void CQReportsWidget::deleteSelectedReports()
{
  const QItemSelectionModel *pSelectionModel = mpTblReports->selectionModel();
  QModelIndexList mappedSelRows;
  size_t i, imax = mpReportDM->rowCount();

  for (i = 0; i < imax; i++)
    {
      if (pSelectionModel->isRowSelected((int) i, QModelIndex()))
        {
          mappedSelRows.append(mpProxyModel->mapToSource(mpProxyModel->index((int) i, 0)));
        }
    }

  if (mappedSelRows.empty())
    {return;}

  mpReportDM->removeRows(mappedSelRows);
}

void CQReportsWidget::slotBtnClearClicked()
{
  int ret = CQMessageBox::question(this, tr("Confirm Delete"), "Delete all Reports?",
                                   QMessageBox::Yes | QMessageBox::No, QMessageBox::No);

  if (ret == QMessageBox::Yes)
    {
      mpReportDM->clear();
    }

  updateDeleteBtns();
}

bool CQReportsWidget::update(ListViews::ObjectType C_UNUSED(objectType), ListViews::Action C_UNUSED(action), const std::string &C_UNUSED(key))
{
  if (!mIgnoreUpdates && isVisible())
    {
      enterProtected();
    }

  return true;
}

bool CQReportsWidget::leave()
{
  return true;
}

bool CQReportsWidget::enterProtected()
{
  if (mpTblReports->selectionModel() != NULL)
    {
      disconnect(mpTblReports->selectionModel(), SIGNAL(selectionChanged(const QItemSelection &, const QItemSelection &)),
                 this, SLOT(slotSelectionChanged(const QItemSelection &, const QItemSelection &)));
    }

  mpProxyModel->setSourceModel(mpReportDM);
  //Set Model for the TableView
  mpTblReports->setModel(NULL);
  mpTblReports->setModel(mpProxyModel);
  connect(mpTblReports->selectionModel(), SIGNAL(selectionChanged(const QItemSelection &, const QItemSelection &)),
          this, SLOT(slotSelectionChanged(const QItemSelection &, const QItemSelection &)));
  updateDeleteBtns();
  mpTblReports->resizeColumnsToContents();
  setFramework(mFramework);
  return true;
}

void CQReportsWidget::updateDeleteBtns()
{
  bool selected = false;
  QModelIndexList selRows = mpTblReports->selectionModel()->selectedRows();

  if (selRows.size() == 0)
    selected = false;
  else
    {
      if (selRows.size() == 1)
        {
          if (mpReportDM->isDefaultRow(mpProxyModel->mapToSource(selRows[0])))
            selected = false;
          else
            selected = true;
        }
      else
        selected = true;
    }

  mpBtnDelete->setEnabled(selected);

  if (mpProxyModel->rowCount() - 1)
    mpBtnClear->setEnabled(true);
  else
    mpBtnClear->setEnabled(false);
}

void CQReportsWidget::slotSelectionChanged(const QItemSelection &C_UNUSED(selected),
    const QItemSelection &C_UNUSED(deselected))
{
  updateDeleteBtns();
}

void CQReportsWidget::dataChanged(const QModelIndex &C_UNUSED(topLeft),
                                  const QModelIndex &C_UNUSED(bottomRight))
{
  mpTblReports->resizeColumnsToContents();
  setFramework(mFramework);
  updateDeleteBtns();
}

void CQReportsWidget::slotDoubleClicked(const QModelIndex proxyIndex)
{
  QModelIndex index = mpProxyModel->mapToSource(proxyIndex);

  if (index.row() < 0)
    return;

  if (mpReportDM->isDefaultRow(index))
    {
      slotBtnNewClicked();
    }

  assert(mpDataModel != NULL);

  if (!mpDataModel->getModel())
    return;

  std::string key = mpDataModel->getReportDefinitionList()->operator[](index.row()).getKey();

  if (CRootContainer::getKeyFactory()->get(key))
    mpListView->switchToOtherWidget(C_INVALID_INDEX, key);
}

void CQReportsWidget::keyPressEvent(QKeyEvent *ev)
{
  if (ev->key() == Qt::Key_Delete)
    slotBtnDeleteClicked();
  else if (ev->key() == Qt::Key_C && (ev->modifiers() & Qt::ControlModifier))
    {
      QModelIndexList selRows = mpTblReports->selectionModel()->selectedRows(0);

      if (selRows.empty())
        {return;}

      QString str;
      QModelIndexList::const_iterator i;

      for (i = selRows.begin(); i != selRows.end(); ++i)
        {
          for (int x = 0; x < mpReportDM->columnCount(); ++x)
            {
              if (!mpTblReports->isColumnHidden(x))
                {
                  if (!str.isEmpty())
                    str += "\t";

                  str += mpReportDM->index(mpProxyModel->mapToSource(*i).row(), x).data().toString();
                }
            }

          str += "\n";
        }

      QApplication::clipboard()->setText(str);
    }
}

void CQReportsWidget::slotFilterChanged()
{
  QRegExp regExp(mpLEFilter->text() + "|New Report", Qt::CaseInsensitive, QRegExp::RegExp);
  mpProxyModel->setFilterRegExp(regExp);
}

void CQReportsWidget::setFramework(int framework)
{
  CopasiWidget::setFramework(framework);
}
